/* File: sound.c

	MZ-2000/2200/80B/B2 Emulator "emz2000" Plugins
		Sound for MZ-1E18 Plugin (QD)

	Copyright (C) 2002-2015 Toshio Fukui
	This file is part of the emz2000 software.
	See copyright notice in the COPYING file.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>
#include <unistd.h>
#include <unistd.h>
#include <esd.h>
#include <limits.h>

#include "common.h"
#include "wavefile.h"

#include "../interface/plugin_posix.h"
#define WORD	unsigned short
#define LPCSTR	const char *
#include "mz1e18.h"

#define DEFAULT_SOUND_VOLUME	((256 - 1) * 7 / 10)	/* Effect Waves (ESD_VOLUME_BASE=256) */
#define module_name		"mz1e18.so"

/* for esd */
static int esd_available = FALSE;
static int esd_sock = - 1;
static esd_info_t *esd_ctrl = NULL;
static int sound_volume;

/* for wavefiles */
typedef struct {
	const char *fn;
	int id;
	int device_type;	/* 3 = QD / 0 = others */
} SOUND_WAVEDATA;

static SOUND_WAVEDATA sound_wavedata[] = {
	{"qdeject.wav",		-1,	3},
	{"qdseek.wav",		-1,	3},
	{"qdset.wav",		-1,	3}
};

int sound_playwav_file( SOUND_WAVLIST no )
{
	if (no < 0 || no >= sizeof(sound_wavedata)/sizeof(SOUND_WAVEDATA))
		return -1;
	if (sound_wavedata[no].id < 0)
		return -1;
	esd_sample_play( esd_sock, sound_wavedata[no].id );
	return 0;
}

static char *sound_loadwav_file( WAVEDATA_STRUCT *ws_p, const char *fn )
{
	int r;
	unsigned long len;
	char *p, *rp;

	r = wavefile_ropen( WAVEFILE_ATTR_DEFAULT, ws_p, fn );
#ifdef EMZ2000_PLGDIR
	if (r < 0) {
		char tpath[PATH_MAX];

		strncpy( tpath, EMZ2000_PLGDIR "/", PATH_MAX );
		tpath[PATH_MAX - 1] = '\0';
		strncat( tpath, fn, PATH_MAX - strlen(tpath) - 1 );
		r = wavefile_ropen( WAVEFILE_ATTR_DEFAULT, ws_p, tpath );
	}
#endif
#ifdef EMZ2000_LIBDIR
	if (r < 0) {
		char tpath[PATH_MAX];

		strncpy( tpath, EMZ2000_LIBDIR "/", PATH_MAX );
		tpath[PATH_MAX - 1] = '\0';
		strncat( tpath, fn, PATH_MAX - strlen(tpath) - 1 );
		r = wavefile_ropen( WAVEFILE_ATTR_DEFAULT, ws_p, tpath );
	}
#endif
	if (r < 0) {
		fprintf( stderr, "%s: read open error (result=%d) \"%s\"\n", module_name, r, fn );
		return NULL;
	}
	len = wavedata_getlenbytes( ws_p );
	p = MALLOC( len );
	if (!p) {
		fprintf( stderr, "%s: memory allocation error (%lu bytes).\n", module_name, len );
		wavefile_rclose( ws_p );
		return NULL;
	}
	rp = p;
	while (len > 0) {
		r = read( wavedata_getfd(ws_p), p, len );
		if (r > 0) {
			len -= r;
			p += r;
		}
		if (r < 0) {
			fprintf( stderr, "%s: read error r=%d (%s)\n", module_name, r, strerror(errno) );
			wavefile_rclose( ws_p );
			return NULL;
		}
	}
	r = wavefile_rclose( ws_p );
	if (r < 0) {
		fprintf( stderr, "%s: read close error\n", module_name );
		return NULL;
	}
	return rp;
}

static void sound_freewav( void )
{
	int i;

	for (i = 0; i < sizeof(sound_wavedata)/sizeof(SOUND_WAVEDATA); i++) {
		if (sound_wavedata[i].id >= 0) {
			esd_sample_free( esd_sock, sound_wavedata[i].id );
			sound_wavedata[i].id = -1;
		}
	}
	return;
}

static int sound_loadwav( void )
{
	int r, i, len, id, trate;
	char *p, esdname[ESD_NAME_MAX];
	WAVEDATA_STRUCT ws;
	esd_format_t tformat;

	for (i = 0; i < sizeof(sound_wavedata)/sizeof(SOUND_WAVEDATA); i++) {
		_TRACE("%d: %s\n", i, sound_wavedata[i].fn);
		wavedata_initstruct( &ws );
		p = sound_loadwav_file( &ws, sound_wavedata[i].fn );
		if (!p)
			continue;
		_TRACE("f=%d ch=%d bytes=%d\n", ws.f, ws.ch, ws.bytes ); 
		tformat = ESD_STREAM | ESD_PLAY;
		if (ws.ch == 1)
			tformat |= ESD_MONO;
		else
			tformat |= ESD_STEREO;
		if (ws.bytes == 1)
			tformat |= ESD_BITS8;
		else
			tformat |= ESD_BITS16;
		trate = ws.f;
		len = wavedata_getlenbytes( &ws );
		strncpy( esdname, sound_wavedata[i].fn, ESD_NAME_MAX );
		esdname[ESD_NAME_MAX - 1] = '\0';
		id = esd_sample_cache( esd_sock, tformat, trate, len, esdname );
		if (id >= 0) {
			r = write( esd_sock, p, len );
			if (r < 0)
				fprintf( stderr, "Wav file \"%s\" read failed.\n", esdname );
			esd_confirm_sample_cache( esd_sock );
			esd_set_stream_pan( esd_sock, id, sound_volume, sound_volume );
			sound_wavedata[i].id = id;
			_TRACE("ok:id=%d\n", id);
		} else
			FREE( p );
	}
	return 0;
}

int sound_enablechk( void )
{
	return esd_available;
}

int sound_updateparams( void )
{
	sound_exit();
	return sound_init();
}

void sound_exit( void )
{
	sound_freewav();
	if (esd_ctrl) {
		esd_free_all_info( esd_ctrl );
		esd_ctrl = NULL;
	}
	if (esd_sock >= 0) {
		close(esd_sock);
		esd_sock = -1;
	}
	esd_available = FALSE;
	return;
}

int sound_init( void )
{
	sound_volume = DEFAULT_SOUND_VOLUME;
	esd_sock = esd_open_sound( NULL /* host */ );
	if (esd_sock < 0) {
		esd_available = FALSE;
		return FALSE;
	}
	esd_ctrl = esd_get_all_info( esd_sock );
	if (esd_ctrl) {
		if (esd_ctrl -> player_list)
			esd_set_stream_pan( esd_sock, esd_ctrl -> player_list -> source_id,
				sound_volume, sound_volume );
		else
			esd_set_stream_pan( esd_sock, 0,
				sound_volume, sound_volume );
		/* esd_print_all_info( esd_ctrl ); */
		esd_free_all_info( esd_ctrl );
		esd_ctrl = NULL;
	}
	esd_available = TRUE;
	sound_loadwav();
	return TRUE;
}

/*
	Local Variables:
	mode:c++
	c-set-style:"k&r"
	c-basic-offset:8
	tab-width:8
	End:
*/
